#ifndef MSP_DATAFILE_COLLECTION_H_
#define MSP_DATAFILE_COLLECTION_H_

#include <msp/core/maputils.h>
#include <msp/core/meta.h>
#include <msp/core/refptr.h>
#include "loader.h"

namespace Msp {
namespace DataFile {

/**
Helper struct to determine whether a Loader has a Collection typedef.
*/
template<typename T>
struct NeedsCollection
{
	struct Yes { char c[2]; };
	struct No { char c; };

	template<typename U>
	static Yes f(typename U::Collection *);
	template<typename U>
	static No f(...);

	enum { value = (sizeof(f<T>(0))==sizeof(Yes)) };
};

class CollectionItemTypeBase;

template<typename T>
class CollectionItemType;

/**
A collection of objects that can be loaded from a datafile.  Each object is
identified by a name, which must be unique across the entire collection.

While this class can be instantiated by itself and used for storing objects,
loading requires that a subclass defines the supported types.  See the add_type
method for details.

Collections also support a notion of "future objects".  These are objects which
are known to be possible to load, but loading them is deferred to the first
time they are requested.
*/
class Collection
{
public:
	/**
	Loads objects into a Collection.  Automatically picks up keywords from
	defined item types.
	*/
	class Loader: public DataFile::Loader
	{
		template<typename T> friend class CollectionItemType;

	private:
		template<typename T, typename S, bool = NeedsCollection<typename T::Loader>::value>
		struct Add;

		Collection &coll;

	public:
		Loader(Collection &);
		Collection &get_object() const { return coll; }
	private:
		template<typename T, typename S, typename C>
		void coll_item(const std::string &n)
		{
			RefPtr<T> it = new T;
			load_sub(*it, dynamic_cast<C &>(coll));
			coll.add<S>(n, it.get());
			it.release();
		}

		template<typename T, typename S>
		void item(const std::string &n)
		{
			RefPtr<T> it = new T;
			load_sub(*it);
			coll.add<S>(n, it.get());
			it.release();
		}
	};

private:
	typedef std::map<std::string, Variant> ItemMap;
	typedef std::list<CollectionItemTypeBase *> TypeList;

	TypeList types;
	ItemMap items;

	Collection(const Collection &);
	Collection &operator=(const Collection &);
public:
	Collection() { }
	virtual ~Collection();

	/** Adds an object into the collection.  The name must not pre-exist.  The
	collection takes ownership of the object. */
	template<typename T>
	void add(const std::string &name, T *item)
	{
		if(!item)
			throw std::invalid_argument("Collection::add(item)");

		typedef RefPtr<typename RemoveConst<T>::Type> RPNCT;

		ItemMap::iterator i = items.find(name);
		if(i!=items.end())
		{
			if(i->second.check_type<RPNCT>())
			{
				// Replace a future object placeholder
				RPNCT &ptr = i->second.value<RPNCT>();
				if(!ptr)
				{
					ptr = item;
					return;
				}
			}

			throw key_error(typeid(ItemMap));
		}

		items.insert(ItemMap::value_type(name, RPNCT(item)));
	}

protected:
	/** Adds the name of a future object to the collection.  The object itself
	will be loaded on first access.  The calling subclass should be prepared to
	create the object on request. */
	template<typename T>
	void add_future(const std::string &name)
	{
		RefPtr<typename RemoveConst<T>::Type> ptr(0);
		insert_unique(items, name, ptr);
	}

	void add_future(const std::string &name);

public:
	/// Gets a typed object from the collection.
	template<typename T>
	T &get(const std::string &name) const
	{
		typedef typename RemoveConst<T>::Type NCT;

		T *ptr = get_item(items, name).value<RefPtr<NCT> >().get();
		if(!ptr)
			throw key_error(typeid(ItemMap));
		return *ptr;
	}

	/** Gets a typed object from the collection.  If the name is not found in
	and a creator for the item type is defined, it is invoked. */
	template<typename T>
	T &get(const std::string &);

private:
	template<typename T>
	void collect_items(std::list<T *> *objects, std::list<std::string> *names, std::list<std::string> *future_names) const
	{
		typedef RefPtr<typename RemoveConst<T>::Type> RPNCT;

		for(ItemMap::const_iterator i=items.begin(); i!=items.end(); ++i)
			if(i->second.check_type<RPNCT>())
			{
				T *ptr = i->second.value<RPNCT>().get();
				if(ptr)
				{
					if(objects)
						objects->push_back(ptr);
					if(names)
						names->push_back(i->first);
				}
				else if(future_names)
					future_names->push_back(i->first);
			}
	}

public:
	/** Returns a list of the names of loaded objects of one type in the
	collection. */
	template<typename T>
	std::list<std::string> get_names() const
	{
		std::list<std::string> result;
		collect_items<T>(0, &result, 0);
		return result;
	}

	/** Returns a list of the names of objects of one type in the collection,
	including any future objects. */
	template<typename T>
	std::list<std::string> get_names()
	{
		std::list<std::string> result;
		collect_items<T>(0, &result, &result);
		return result;
	}

	/// Returns a list of loaded objects of one type in the collection.
	template<typename T>
	std::list<T *> get_list() const
	{
		std::list<T *> result;
		collect_items<T>(&result, 0, 0);
		return result;
	}

	/** Returns a list of objects of one type in the collection.  Any future
	objects of that type are loaded and returned in the list. */
	template<typename T>
	std::list<T *> get_list()
	{
		std::list<T *> result;
		std::list<std::string> future;
		collect_items<T>(&result, 0, &future);
		for(std::list<std::string>::iterator i=future.begin(); i!=future.end(); ++i)
			result.push_back(&get<T>(*i));
		return result;
	}

private:
	template<typename T>
	unsigned get_status(const std::string &name) const
	{
		ItemMap::const_iterator i = items.find(name);
		if(i==items.end())
			return false;

		typedef RefPtr<typename RemoveConst<T>::Type> RPNCT;
		if(!i->second.check_type<RPNCT>())
			return false;

		T *ptr = i->second.value<RPNCT>().get();
		return ptr ? 1 : 2;
	}

public:
	/// Checks whether a typed object exists and is loaded in the collection.
	template<typename T>
	bool contains(const std::string &name) const
	{ return get_status<T>(name)==1; }

	/** Checks whether a typed object exists in the collection, as either a
	loaded or future object. */
	template<typename T>
	bool contains(const std::string &name)
	{ return get_status<T>(name)>0; }

	/// Returns the name of an item in the collection.
	template<typename T>
	const std::string &get_name(T *d) const
	{
		typedef RefPtr<typename RemoveConst<T>::Type> RPNCT;

		for(ItemMap::const_iterator i=items.begin(); i!=items.end(); ++i)
			if(i->second.check_type<RPNCT>())
				if(i->second.value<RPNCT>().get()==d)
					return i->first;
	
		// XXX Need better exception class
		throw std::runtime_error("Item not found in collection");
	}

protected:
	/** Adds a type to the collection.  The returned descriptor object reference
	can be used to define how objects of that type can be loaded. */
	template<typename T>
	CollectionItemType<T> &add_type();
};


template<typename T, typename S>
struct Collection::Loader::Add<T, S, false>
{
	static void add(Loader &loader, const std::string &kwd)
	{ loader.add(kwd, &Loader::item<T, S>); }
};

template<typename T, typename S>
struct Collection::Loader::Add<T, S, true>
{
	static void add(Loader &loader, const std::string &kwd)
	{ loader.add(kwd, &Loader::coll_item<T, S, typename T::Loader::Collection>); }
};


class CollectionItemTypeBase
{
protected:
	class TagBase
	{
	protected:
		TagBase() { }
	public:
		virtual ~TagBase() { }
	};

	template<typename T>
	class Tag: public TagBase
	{ };

	std::string kwd;
	std::vector<std::string> suffixes;
	TagBase *tag;

	CollectionItemTypeBase();
public:
	virtual ~CollectionItemTypeBase();

	void set_keyword(const std::string &);
	void add_suffix(const std::string &);
	virtual void add_to_loader(Collection::Loader &) const = 0;
	virtual bool can_create() const = 0;
	virtual void create_item(Collection &, const std::string &) const = 0;
	bool match_name(const std::string &) const;
	virtual Variant create_future() const = 0;

	template<typename T>
	bool check_type() const
	{ return dynamic_cast<Tag<T> *>(tag); }
};


/**
Describes a type of item that can be loaded by a Collection.  These are created
by Collection::add_type.
*/
template<typename T>
class CollectionItemType: public CollectionItemTypeBase
{
private:
	class CreatorBase
	{
	protected:
		CreatorBase() { }
	public:
		virtual ~CreatorBase() { }

		virtual T *create(Collection &, const std::string &) const = 0;
	};

	template<typename C>
	class Creator: public CreatorBase
	{
	public:
		typedef T *(C::*FuncPtr)(const std::string &);

	private:
		FuncPtr func;

	public:
		Creator(FuncPtr f): func(f) { }

		virtual T *create(Collection &coll, const std::string &name) const
		{ return (static_cast<C &>(coll).*func)(name); }
	};

	class StoreBase
	{
	protected:
		StoreBase() { }
	public:
		virtual ~StoreBase() { }

		virtual void store(Collection &, const std::string &, T *) = 0;
		virtual Variant create_future() const = 0;

		virtual void add_to_loader(Collection::Loader &, const std::string &) = 0;
	};

	template<typename S>
	class Store: public StoreBase
	{
	public:
		virtual void store(Collection &coll, const std::string &name, T *obj)
		{ coll.add(name, static_cast<S *>(obj)); }

		virtual Variant create_future() const
		{ return RefPtr<S>(0); }

		virtual void add_to_loader(Collection::Loader &loader, const std::string &kwd)
		{ Collection::Loader::Add<T, S>::add(loader, kwd); }
	};

	CreatorBase *creat;
	StoreBase *store;

public:
	CollectionItemType():
		creat(0), store(new Store<T>)
	{ tag = new Tag<T>; }

	~CollectionItemType()
	{
		delete creat;
		delete store;
	}

	/** Sets a datafile keyword for this item type.  The Collection's loader
	will accept a statement with this keyword and a single string argument - the
	item's name. */
	CollectionItemType &keyword(const std::string &k)
	{
		set_keyword(k);
		return *this;
	}

	/** Adds a suffix that is used to match names when looking for future
	objects.  There is no implied separator; a name matches if it ends with the
	suffix.  If a keyword is defined before any suffixes, then "."+keyword is
	added as a suffix. */
	CollectionItemType &suffix(const std::string &s)
	{
		add_suffix(s);
		return *this;
	}

	/** Attaches a creator function to this item type.  If an item is not found
	in the Collection, the creator function for its type is called to create it.
	The function must be a member of the Collection subclass containing the
	type.  It must return the created object, or null if it could not be
	created.  It's also permissible to load the item via other means and then
	return null. */
	template<typename C>
	CollectionItemType &creator(T *(C::*func)(const std::string &))
	{
		delete creat;
		creat = new Creator<C>(func);
		return *this;
	}

	/** Specifies the storage type for items of this type.  It must be a base
	class of the actual type.  */
	template<typename S>
	CollectionItemType &store_as()
	{
		delete tag;
		tag = new Tag<S>;
		delete store;
		store = new Store<S>;
		return *this;
	}

	virtual void add_to_loader(Collection::Loader &loader) const
	{ store->add_to_loader(loader, kwd); }

	virtual bool can_create() const
	{ return creat!=0; }

	virtual void create_item(Collection &coll, const std::string &name) const
	{
		if(!creat)
			throw std::runtime_error("no creator");
		T *obj = creat->create(coll, name);
		if(obj)
			store->store(coll, name, obj);
	}

	virtual Variant create_future() const
	{ return store->create_future(); }
};


template<typename T>
T &Collection::get(const std::string &name)
{
	typedef typename RemoveConst<T>::Type NCT;

	ItemMap::iterator i = items.find(name);
	if(i!=items.end())
	{
		NCT *ptr = i->second.value<RefPtr<NCT> >().get();
		if(ptr)
			return *ptr;
	}

	for(TypeList::iterator j=types.begin(); j!=types.end(); ++j)
		if((*j)->can_create() && (*j)->check_type<NCT>())
			(*j)->create_item(*this, name);

	return *get_item(items, name).value<RefPtr<NCT> >();
}

template<typename T>
CollectionItemType<T> &Collection::add_type()
{
	CollectionItemType<T> *type = new CollectionItemType<T>;
	types.push_back(type);
	return *type;
}

} // namespace DataFile
} // namespace Msp

#endif
